/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpiimpl.h"

#if defined(FINEGRAIN_MPI)
/* Different barrier implementation options below. */
#define NESTED_COMMUNICATOR 0
/* For co-proclets */
#define GS_BARRIER 0 /* Gather-Scatter Barrier */
#define DEFAULT_BARRIER 0 /* Normal barrier using point to point communication for co-proclets */
#define RING_BARRIER 0  /* Twice-ring barrier */
#define SM_BARRIER 1 /* Shared memory counter barrier */
/* For leader proclets in different address spaces */
#define DEFAULT_PID_BARRIER 0 /* Normal barrier using point to point communication for leader proclets */
#define SM_PID_BARRIER 1 /* Shared memory counter barrier for leader proclets */

static int MPIR_Create_nestedcomm( MPI_Comm comm, MPID_Comm **coFGP_comm_dblptr, MPID_Comm **pid_comm_dblptr, int *comsize_ptr, int *myrank_ptr );
static int find_numofcoFGPs_and_pidRepresentative( MPID_Comm *comm_ptr );
#endif

/* -- Begin Profiling Symbol Block for routine MPI_Barrier */
#if defined(HAVE_PRAGMA_WEAK)
#pragma weak MPI_Barrier = PMPI_Barrier
#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#pragma _HP_SECONDARY_DEF PMPI_Barrier  MPI_Barrier
#elif defined(HAVE_PRAGMA_CRI_DUP)
#pragma _CRI duplicate MPI_Barrier as PMPI_Barrier
#endif
/* -- End Profiling Symbol Block */


/* Define MPICH_MPI_FROM_PMPI if weak symbols are not supported to build
   the MPI routines */
#ifndef MPICH_MPI_FROM_PMPI
#undef MPI_Barrier
#define MPI_Barrier PMPI_Barrier


/* This is the default implementation of the barrier operation.  The
   algorithm is:
   
   Algorithm: MPI_Barrier

   We use the dissemination algorithm described in:
   Debra Hensgen, Raphael Finkel, and Udi Manbet, "Two Algorithms for
   Barrier Synchronization," International Journal of Parallel
   Programming, 17(1):1-17, 1988.  

   It uses ceiling(lgp) steps. In step k, 0 <= k <= (ceiling(lgp)-1),
   process i sends to process (i + 2^k) % p and receives from process 
   (i - 2^k + p) % p.

   Possible improvements: 

   End Algorithm: MPI_Barrier

   This is an intracommunicator barrier only!
*/

/* not declared static because it is called in ch3_comm_connect/accept */
int MPIR_Barrier( MPID_Comm *comm_ptr )
{
    static const char FCNAME[] = "MPIR_Barrier";
    int size, rank, src, dst, mask, mpi_errno=MPI_SUCCESS;
    MPI_Comm comm;

    FGP_DBG_ENTER(MPIR_Barrier);
    /*
      HK: 
      In  /home/humaira/kamalphd-svn/software/MPICH2/mpich2-custom/src/mpid/ch3/src/mpid_init.c:
      MPID_Init[213]                 comm->local_size = pg_size;
      MPID_Init[256]                 comm->local_size = 1;
      MPID_Init[292]                 comm->local_size = pg_size;

     */
#if defined(FINEGRAIN_MPI)
    size = comm_ptr->totprocs; /* HK: setting size to totprocs. TODO Double-check. It was local_size.*/
    /* HK: TODO (for NESTED_COMMUNICATOR)
       Use a simpler way to enforce barrier when there is a single HWP
       and avoid unnecessary communication and copy overhead.
       Note: Hierarchical barrier is now already in place when using shared memory leader based algorithm.
     */
#else
    size = comm_ptr->local_size;
#endif
    rank = comm_ptr->rank;
    
    /* Trivial barriers return immediately */
    if (size == 1) return MPI_SUCCESS;

    comm = comm_ptr->handle;

    /* Only one collective operation per communicator can be active at any
       time */
    MPIDU_ERR_CHECK_MULTIPLE_THREADS_ENTER( comm_ptr );

    mask = 0x1;
    while (mask < size) {
        dst = (rank + mask) % size;
        src = (rank - mask + size) % size;
        mpi_errno = MPIC_Sendrecv(NULL, 0, MPI_BYTE, dst,
                                  MPIR_BARRIER_TAG, NULL, 0, MPI_BYTE,
                                  src, MPIR_BARRIER_TAG, comm,
                                  MPI_STATUS_IGNORE);
	/* --BEGIN ERROR HANDLING-- */
        if (mpi_errno)
	{
	    mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**fail", 0);
	    return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
        mask <<= 1;
    }

    FGP_DBG_EXIT(MPIR_Barrier);
    
    MPIDU_ERR_CHECK_MULTIPLE_THREADS_EXIT( comm_ptr );

    return mpi_errno;
}


#if 0

/* This is the default implementation of the barrier operation.  The
   algorithm is:
   
   Algorithm: MPI_Barrier

   Find the largest power of two that is less than or equal to the size of 
   the communicator.  Call tbis twon_within.

   Divide the communicator by rank into two groups: those with 
   rank < twon_within and those with greater rank.  The barrier
   executes in three steps.  First, the group with rank >= twon_within
   sends to the first (size-twon_within) ranks of the first group.
   That group then executes a recursive doubling algorithm for the barrier.
   For the third step, the first (size-twon_within) ranks send to the top
   group.  This is the same algorithm used in MPICH-1.

   Possible improvements: 
   The upper group could apply recursively this approach to reduce the 
   total number of messages sent (in the case of of a size of 2^n-1, there 
   are 2^(n-1) messages sent in the first and third steps).

   End Algorithm: MPI_Barrier

   This is an intracommunicator barrier only!
*/
int MPIR_Barrier( MPID_Comm *comm_ptr )
{
    int size, rank;
    int twon_within, n2, remaining, gap, partner;
    MPID_Request *request_ptr;
    int mpi_errno = MPI_SUCCESS;
    
    size = comm_ptr->remote_size;
    rank = comm_ptr->rank;

    /* Trivial barriers return immediately */
    if (size == 1) return MPI_SUCCESS;

    /* Only one collective operation per communicator can be active at any
       time */
    MPIDU_ERR_CHECK_MULTIPLE_THREADS_ENTER( comm_ptr );
    
    /* Find the twon_within (this could be cached if more routines
     need it) */
    twon_within = 1;
    n2          = 2;
    while (n2 <= size) { twon_within = n2; n2 <<= 1; }
    remaining = size - twon_within;

    if (rank < twon_within) {
	/* First step: receive from the upper group */
	if (rank < remaining) {
	    MPID_Recv( 0, 0, MPI_BYTE, twon_within + rank, MPIR_BARRIER_TAG, 
		       comm_ptr, MPID_CONTEXT_INTRA_COLL, MPI_STATUS_IGNORE,
		       &request_ptr );
	    if (request_ptr) {
		mpi_errno = MPIC_Wait(request_ptr);
		MPID_Request_release(request_ptr);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
		    goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
	    }
	}
	/* Second step: recursive doubling exchange */
	for (gap=1; gap<twon_within; gap <<= 1) {
	    partner = (rank ^ gap);
	    MPIC_Sendrecv( 0, 0, MPI_BYTE, partner, MPIR_BARRIER_TAG,
			   0, 0, MPI_BYTE, partner, MPIR_BARRIER_TAG,
			   comm_ptr->handle, MPI_STATUS_IGNORE );
	}

	/* Third step: send to the upper group */
	if (rank < remaining) {
	    MPID_Send( 0, 0, MPI_BYTE, rank + twon_within, MPIR_BARRIER_TAG,
		       comm_ptr, MPID_CONTEXT_INTRA_COLL, &request_ptr );
	    if (request_ptr) {
		mpi_errno = MPIC_Wait(request_ptr);
		MPID_Request_release(request_ptr);
		/* --BEGIN ERROR HANDLING-- */
		if (mpi_errno != MPI_SUCCESS)
		{
		    goto fn_exit;
		}
		/* --END ERROR HANDLING-- */
	    }
	}
    }
    else {
	/* For the upper group, step one is a send */
	MPID_Send( 0, 0, MPI_BYTE, rank - twon_within, MPIR_BARRIER_TAG,
		   comm_ptr, MPID_CONTEXT_INTRA_COLL, &request_ptr );
	if (request_ptr) {
	    mpi_errno = MPIC_Wait(request_ptr);
	    MPID_Request_release(request_ptr);
	    /* --BEGIN ERROR HANDLING-- */
	    if (mpi_errno != MPI_SUCCESS)
	    {
		goto fn_exit;
	    }
	    /* --END ERROR HANDLING-- */
	}
	/* There is no second step; for the third step, recv */
	MPID_Recv( 0, 0, MPI_BYTE, rank - twon_within, MPIR_BARRIER_TAG, 
		   comm_ptr, MPID_CONTEXT_INTRA_COLL, MPI_STATUS_IGNORE,
		   &request_ptr );
	if (request_ptr) {
	    mpi_errno = MPIC_Wait(request_ptr);
	    MPID_Request_release(request_ptr);
	    /* --BEGIN ERROR HANDLING-- */
	    if (mpi_errno != MPI_SUCCESS)
	    {
		goto fn_exit;
	    }
	    /* --END ERROR HANDLING-- */
	}
    }

  fn_exit:
    MPIDU_ERR_CHECK_MULTIPLE_THREADS_EXIT( comm_ptr );

    return mpi_errno;
}
#endif


/* not declared static because a machine-specific function may call this one 
   in some cases */
int MPIR_Barrier_inter( MPID_Comm *comm_ptr )
{
    static const char FCNAME[] = "MPIR_Barrier_inter";
    int rank, mpi_errno, i, root;
    MPID_Comm *newcomm_ptr = NULL;

    rank = comm_ptr->rank;

    /* Get the local intracommunicator */
    if (!comm_ptr->local_comm)
	MPIR_Setup_intercomm_localcomm( comm_ptr );

    newcomm_ptr = comm_ptr->local_comm;

    /* do a barrier on the local intracommunicator */
    mpi_errno = MPIR_Barrier(newcomm_ptr);
    /* --BEGIN ERROR HANDLING-- */
    if (mpi_errno)
    {
	mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**fail", 0);
	return mpi_errno;
    }
    /* --END ERROR HANDLING-- */

    /* rank 0 on each group does an intercommunicator broadcast to the
       remote group to indicate that all processes in the local group
       have reached the barrier. We do a 1-byte bcast because a 0-byte
       bcast will just return without doing anything. */
    
    /* first broadcast from left to right group, then from right to
       left group */
    if (comm_ptr->is_low_group) {
        /* bcast to right*/
        root = (rank == 0) ? MPI_ROOT : MPI_PROC_NULL;
        mpi_errno = MPIR_Bcast_inter(&i, 1, MPI_BYTE, root, comm_ptr); 
	/* --BEGIN ERROR HANDLING-- */
        if (mpi_errno)
	{
	    mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**fail", 0);
	    return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
        /* receive bcast from right */
        root = 0;
        mpi_errno = MPIR_Bcast_inter(&i, 1, MPI_BYTE, root, comm_ptr); 
	/* --BEGIN ERROR HANDLING-- */
        if (mpi_errno)
	{
	    mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**fail", 0);
	    return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
    }
    else {
        /* receive bcast from left */
        root = 0;
        mpi_errno = MPIR_Bcast_inter(&i, 1, MPI_BYTE, root, comm_ptr); 
	/* --BEGIN ERROR HANDLING-- */
        if (mpi_errno)
	{
	    mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**fail", 0);
	    return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
        /* bcast to left */
        root = (rank == 0) ? MPI_ROOT : MPI_PROC_NULL;
        mpi_errno = MPIR_Bcast_inter(&i, 1, MPI_BYTE, root, comm_ptr);  
	/* --BEGIN ERROR HANDLING-- */
        if (mpi_errno)
	{
	    mpi_errno = MPIR_Err_create_code(mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, "**fail", 0);
	    return mpi_errno;
	}
	/* --END ERROR HANDLING-- */
    }

    return mpi_errno;
}



#endif

#undef FUNCNAME
#define FUNCNAME MPI_Barrier

/*@

MPI_Barrier - Blocks until all processes in the communicator have
reached this routine.  

Input Parameter:
. comm - communicator (handle) 

Notes:
Blocks the caller until all processes in the communicator have called it; 
that is, the call returns at any process only after all members of the
communicator have entered the call.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
@*/
int MPI_Barrier( MPI_Comm comm )
{
    static const char FCNAME[] = "MPI_Barrier";
    int mpi_errno = MPI_SUCCESS;
    MPID_Comm *comm_ptr = NULL;
    MPID_MPI_STATE_DECL(MPID_STATE_MPI_BARRIER);

    MPIR_ERRTEST_INITIALIZED_ORDIE();
    
    MPIU_THREAD_SINGLE_CS_ENTER("coll");
    MPID_MPI_COLL_FUNC_ENTER(MPID_STATE_MPI_BARRIER);
    
    /* Validate parameters, especially handles needing to be converted */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
	    MPIR_ERRTEST_COMM(comm, mpi_errno);
            if (mpi_errno != MPI_SUCCESS) goto fn_fail;
	}
        MPID_END_ERROR_CHECKS;
    }
#   endif /* HAVE_ERROR_CHECKING */

    /* Convert MPI object handles to object pointers */
    MPID_Comm_get_ptr( comm, comm_ptr );
    
    /* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
	    /* Validate communicator */
            MPID_Comm_valid_ptr( comm_ptr, mpi_errno );
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
    }
#   endif /* HAVE_ERROR_CHECKING */

    /* ... body of routine ...  */
    
#if !defined(FINEGRAIN_MPI) 
    if (comm_ptr->coll_fns != NULL && comm_ptr->coll_fns->Barrier != NULL)
    {
	mpi_errno = comm_ptr->coll_fns->Barrier(comm_ptr); 
    }
    else
#endif
    {
	MPIU_THREADPRIV_DECL;
	MPIU_THREADPRIV_GET;
        MPIR_Nest_incr();
        if (comm_ptr->comm_kind == MPID_INTRACOMM) {
            
#if defined(FINEGRAIN_MPI) 
#if NESTED_COMMUNICATOR /* HK: Creating coFGP_comm and pid_comm, if they don't already exist */
            int comsize = -1, myrank = -1;
            MPID_Comm *coFGP_comm_ptr = NULL, *pid_comm_ptr = NULL;
            mpi_errno = MPIR_Create_nestedcomm(comm, &coFGP_comm_ptr, &pid_comm_ptr, &comsize, &myrank);
            if (mpi_errno) goto fn_fail;
#endif             
            /***************************************************************/
 #if DEFAULT_BARRIER && NESTED_COMMUNICATOR     /* First part of MPIR_Barrier on coFGP_comm_ptr */   
            mpi_errno = MPIR_Barrier(coFGP_comm_ptr); /* HK: This is much more expensive than a ring barrier */
 
            
            /***************************************************************/
 #elif GS_BARRIER && NESTED_COMMUNICATOR
            /* HK: Gather/ Scatter barrier on coFGP_comm_ptr. Does not scale as well as twice-ring */
            int *recvbuf = MPIU_Malloc(sizeof(int)*comsize);
            int *sendbuf = MPIU_Malloc(sizeof(int)*comsize);
            memset(sendbuf, 1, sizeof(int)*comsize);
            int root = 0;
            MPI_Gather(sendbuf, 1, MPI_INT, recvbuf, 1, MPI_INT, root, *(comm_ptr->coFGP_comm));
            /* end of part-1 of GS_BARRIER */

            
            /***************************************************************/
 #elif RING_BARRIER && NESTED_COMMUNICATOR
            /* HK: Sending in a ring of coFGPs */
            int sendtoken = 1, recvtoken;
            MPI_Status status;            
            int mytag = 1212; /* HK: Does this have to be unique and not used at user level? It seems any tag values should be fine because *(comm_ptr->coFGP_comm) is not being used at user level and its context_id should be unique */
            if(comsize > 1){ /* HK: No need to do local coFGP barrier if comsize == 1 */
                if (comm_ptr->isRepresentative){
                    MPI_Send(&sendtoken, 1, MPI_INT, (myrank+1)/*receiver rank*/, mytag, *(comm_ptr->coFGP_comm));
                    MPI_Recv(&recvtoken, 1, MPI_INT, (comsize-1)/*sender rank*/, mytag, *(comm_ptr->coFGP_comm), &status);
                }
                else{
                    MPI_Recv(&recvtoken, 1, MPI_INT, (myrank-1)/*sender rank*/, mytag, *(comm_ptr->coFGP_comm), &status);
                    MPI_Send(&sendtoken, 1, MPI_INT, ((myrank+1)%comsize)/*receiver rank*/, mytag, *(comm_ptr->coFGP_comm));
                }
            }
            /* end of part-1 of RING_BARRIER */

            
            /***************************************************************/
 #elif SM_BARRIER && !NESTED_COMMUNICATOR

            find_numofcoFGPs_and_pidRepresentative( comm_ptr );
            /* Shared memory counter barrier.
               HK: Concurrent barriers can now be called by subsets of proclets in same address space.
            */
            //            int coproclet_size = coFGP_comm_ptr->totprocs;
            int coproclet_size = comm_ptr->numofcoFGPs;
            assert( (comm_ptr->co_shared_vars != NULL) && (comm_ptr->co_shared_vars->co_barrier_vars != NULL) );
            int proclet_sense = comm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal;

            /* SM barrier using leader signal. This would avoid doing a full
               fledged SM barrier again after the SM pid barrier. */
            if(comm_ptr->numofcoFGPs > 1){ /* HK: Do barrier only if there are more than one collocated FGPs in this pid */
                if(!comm_ptr->isRepresentative){ /* non-leader proclets */
                    (comm_ptr->co_shared_vars->co_barrier_vars->coproclet_counter)++;
                    if(comm_ptr->co_shared_vars->co_barrier_vars->coproclet_counter == (coproclet_size-1)){ /* excluding the leader/representative */
                        comm_ptr->co_shared_vars->co_barrier_vars->leader_signal = 1;
                    }
                    while(comm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal == proclet_sense)
                        FG_Yield();   
                }
                else{ /* leader proclet */
                    while(comm_ptr->co_shared_vars->co_barrier_vars->leader_signal == 0)
                        FG_Yield();
                }
            }

 #else
            MPIU_Assert(0); /* Should not be here! No valid options were selected */
           
            
 #endif  /* end of part-1 of co-proclets barrier */

            
            /*-------------------------------------------------------------*/
 #if DEFAULT_PID_BARRIER && NESTED_COMMUNICATOR
            /* HK: Barrier on pid_comm */
            if (comm_ptr->isRepresentative){
                mpi_errno = MPIR_Barrier(pid_comm_ptr); /* HK: If pid_comm_ptr->totprocs is 1, it returns immediately. */
            }
 
            
            /*-------------------------------------------------------------*/
 #elif SM_PID_BARRIER
            /* HK: Shared memory barrier only. From ch3i_comm.c */
            if (comm_ptr->isRepresentative){
#if NESTED_COMMUNICATOR
                /*
                  HK:
                  OLD implementation that uses nested communicator
                  THIS IS ONLY FOR THE CASE WHEN ALL HWPs
                  ARE ON SAME PHYSICAL NODE 
                  There are only local procs -- do shared memory barrier only
                */

                if(NULL == pid_comm_ptr->barrier_vars){
                    int i;
                    int c;
                    
                    for (i = 0; i < MPID_NEM_NUM_BARRIER_VARS; ++i) {
                        c = MPID_NEM_CAS_INT (&MPID_nem_mem_region.barrier_vars[i].context_id, -1/*NULL_CONTEXT_ID*/, pid_comm_ptr->context_id);
                        if (c == -1/*NULL_CONTEXT_ID*/ || c == pid_comm_ptr->context_id)
                            {
                                pid_comm_ptr->barrier_vars = &(MPID_nem_mem_region.barrier_vars[i]);
                                MPID_NEM_WRITE_BARRIER();
                                MPID_NEM_ATOMIC_INC (&((pid_comm_ptr->barrier_vars)->usage_cnt));
                                break;
                            }
                    }
                    assert(NULL != pid_comm_ptr->barrier_vars);                    
                }
                volatile MPID_nem_barrier_vars_t *barrier_vars;
                barrier_vars = pid_comm_ptr->barrier_vars;
                int local_size = pid_comm_ptr->totprocs; /* This the number of local procs on same physical node. */                
                int prev;
                int sense;
                
                sense = barrier_vars->sig;
                //MPID_NEM_READ_BARRIER();
                OPA_read_barrier();
                prev = MPID_NEM_FETCH_AND_INC (&barrier_vars->cnt);

                if (prev == (local_size-1)){
                    barrier_vars->cnt = 0;
                    //MPID_NEM_WRITE_BARRIER();
                    OPA_write_barrier();
                    barrier_vars->sig = 1 - sense;
                }
                else{
                    while (barrier_vars->sig == sense){
                        FG_Yield();
                    }
                }
                        
#else
                /*
                  HK: In src/mpid/ch3/channels/nemesis/src/ch3i_comm.c
                  MPIDI_CH3I_comm_create()    comm->coll_fns = &collective_functions;
                  In /src/mpid/ch3/channels/nemesis/src/ch3i_comm.c, coll_fns->Barrier
                  refers to a local static function called barrier.
                  Note: MPIDI_CH3I_comm_create() is called by MPID_Dev_comm_create_hook()
                  MPIDI_CH3I_comm_create() calls find_local_and_external().
                */
                mpi_errno = comm_ptr->coll_fns->Barrier(comm_ptr);
                if (mpi_errno) goto fn_fail;
#endif                                     
            } /* closing brace of if(comm_ptr->isRepresentative) */
  #else
            MPIU_Assert(0); /* Should not be here! No valid options were selected */

 #endif  /* end of leader-proclets barrier  */


            /***************************************************************/
 #if DEFAULT_BARRIER && NESTED_COMMUNICATOR  /* Remaining part of MPIR_Barrier on coFGP_comm_ptr */   
            /* HK:
               If calling MPIR_Barrier on coFGP_comm_ptr, instead of twice-ring barrier:
               For cases like -nfg X -n 1, where X>=1, don't call Barrier on coFGP_comm twice.
               For twice-ring barrier, we still need to do an second ring because once around
               the ring does not constitute a barrier.
            */
            
            if(comm_ptr->local_size > 1){ /* HK: There is more than one HWPs */
                mpi_errno = MPIR_Barrier(coFGP_comm_ptr); /* HK: This is much more expensive than a ring barrier */
            }
 
            
            /***************************************************************/
 #elif GS_BARRIER && NESTED_COMMUNICATOR /* Remaining part of Gather/Scatter Barrier on coFGP_comm_ptr */   
            MPI_Scatter(sendbuf, 1, MPI_INT, recvbuf, 1, MPI_INT, root, *(comm_ptr->coFGP_comm));
            MPIU_Free(recvbuf);
            MPIU_Free(sendbuf);
 

            /***************************************************************/
 #elif RING_BARRIER && NESTED_COMMUNICATOR  /* Remaining part of Twice-Ring Barrier */   
            if(comsize > 1){
                mytag = 1213;
                if (comm_ptr->isRepresentative){
                    MPI_Send(&sendtoken, 1, MPI_INT, (myrank+1)/*receiver rank*/, mytag, *(comm_ptr->coFGP_comm));
                    MPI_Recv(&recvtoken, 1, MPI_INT, (comsize-1)/*sender rank*/, mytag, *(comm_ptr->coFGP_comm), &status);
                }
                else{
                    MPI_Recv(&recvtoken, 1, MPI_INT, (myrank-1)/*sender rank*/, mytag, *(comm_ptr->coFGP_comm), &status);
                    MPI_Send(&sendtoken, 1, MPI_INT, ((myrank+1)%comsize)/*receiver rank*/, mytag, *(comm_ptr->coFGP_comm));
                }
            }
 
            
            /***************************************************************/
 #elif SM_BARRIER && !NESTED_COMMUNICATOR
            /* Shared memory counter barrier using leader signal.*/ 
            if(comm_ptr->numofcoFGPs > 1){ /* HK: Do barrier only if there are more than one collocated FGPs in this pid */
                if(comm_ptr->isRepresentative){
                    comm_ptr->co_shared_vars->co_barrier_vars->leader_signal = 0;
                    comm_ptr->co_shared_vars->co_barrier_vars->coproclet_counter = 0;
                    comm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal = 1 - comm_ptr->co_shared_vars->co_barrier_vars->coproclet_signal;
                }
            }

 #else
            MPIU_Assert(0); /* Should not be here! No valid options were selected */
            
 #endif /* end of part-2 of co-proclets barrier */
            
            
#else /* #else case for #if defined(FINEGRAIN_MPI) above. */
	    mpi_errno = MPIR_Barrier( comm_ptr );
#endif
        }
        else {
            /* intercommunicator */ 
            mpi_errno = MPIR_Barrier_inter( comm_ptr );
	}
        MPIR_Nest_decr();
    }

    /* --BEGIN ERROR HANDLING-- */
    if (mpi_errno != MPI_SUCCESS) goto fn_fail;
    /* --END ERROR HANDLING-- */

    /* ... end of body of routine ... */

  fn_exit:
    MPID_MPI_COLL_FUNC_EXIT(MPID_STATE_MPI_BARRIER);
    MPIU_THREAD_SINGLE_CS_EXIT("coll");
    return mpi_errno;

  fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
	    mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
	    "**mpi_barrier", "**mpi_barrier %C", comm);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}



#if defined(FINEGRAIN_MPI)

static int MPIR_Create_nestedcomm( MPI_Comm comm, MPID_Comm **coFGP_comm_dblptr, MPID_Comm **pid_comm_dblptr, int *comsize_ptr, int *myrank_ptr )
{
    static const char FCNAME[] = "MPIR_Create_nestedcomm";
    int mpi_errno = MPI_SUCCESS;
    MPID_Comm *comm_ptr = NULL;
    /* Convert MPI object handles to object pointers */
    MPID_Comm_get_ptr( comm, comm_ptr );

    if( NULL == comm_ptr->coFGP_comm ){
        MPIU_Assert( NULL == comm_ptr->pid_comm );
        comm_ptr->coFGP_comm = MPIU_Malloc(sizeof(MPI_Comm));
        comm_ptr->pid_comm   = MPIU_Malloc(sizeof(MPI_Comm));
        MPI_Group *orig_group = MPIU_Malloc(sizeof(MPI_Group));
        MPI_Group *coFGP_group = MPIU_Malloc(sizeof(MPI_Group));
        MPI_Group *pid_group   = MPIU_Malloc(sizeof(MPI_Group));

        int totfgps = comm_ptr->totprocs; /* HK: number of FGPs that are member of this communicator */
        int totpids = comm_ptr->local_size; /* HK: Total number of HWPs in COMM_WORLD */
        int *coFGParray = MPIU_Malloc(sizeof(int) * totfgps); /*HK: Allocating max size if all the FGPs are in a single pid. Later it is realloc'd */
        int *repPIDarray = MPIU_Malloc(sizeof(int) * totpids); /* HK: pid is the index of this array and value at index is the smallest FGP rank in that pid */
        int *nonRepPIDarray = MPIU_Malloc(sizeof(int) * totfgps); /*HK: All non representative FGPs are stored here, threfore the max size of array is totfgps. Later it is realloc'd */
        int *pidCommarray = NULL; /* HK: This will be used to store the ranks that will be used to create the group associated with pid_comm */
        memset(repPIDarray, -1, sizeof(int) * totpids); /* HK: Initializing to -1 */
        int i, numofcoFGPs=0, numPidsInComm=0, nonRepFGPs=0, isRep=0;

        for (i=0; i<totfgps; i++){
            int pid = -1, worldrank = -1;
            RTWPmapFind(comm_ptr->co_shared_vars->rtw_map, i, &worldrank, &pid);
            if(pid == comm_ptr->p_rank){ /* HK: It is a coFGP if it has the same pid as the calling FGP. _p_rank_ */
                coFGParray[numofcoFGPs] = i;
                numofcoFGPs++;
            }
            FGP_DBG_PRINTF("fgrank in comm=%d, pid=%d\n", i, pid);
            /* Separating the representative and non-representative FGPs */
            if( (repPIDarray[pid] != -1) && (repPIDarray[pid] < i) ){
                nonRepPIDarray[nonRepFGPs] = i;
                nonRepFGPs++;
            }
            if( (repPIDarray[pid] == -1) || (repPIDarray[pid] > i) ){
                if(repPIDarray[pid] != -1){
                    nonRepPIDarray[nonRepFGPs] = repPIDarray[pid];
                    nonRepFGPs++;
                }
                repPIDarray[pid] = i;
            }
        }

        /* HK: Adjusting the size of array */
        coFGParray = MPIU_Realloc(coFGParray, sizeof(int)*numofcoFGPs);
        nonRepPIDarray = MPIU_Realloc(nonRepPIDarray, sizeof(int)*nonRepFGPs);
        MPIU_Assert(coFGParray != NULL); /* HK: coFGParray size must be atleast 1.
                                            nonRepFGPs, however, may be zero,
                                            e.g. -nfg 1 -n 1 */
        MPIU_Assert((totfgps-nonRepFGPs) > 0); /* HK: There must be atleast one
                                                  representative FGP */

        if( repPIDarray[comm_ptr->p_rank] == comm_ptr->rank ){ /* HK: If the calling
                                                                  FGP is the smallest rank
                                                                  in my pid. _p_rank_ */
            isRep = 1;
            comm_ptr->isRepresentative = 1;
        }
        else{
            isRep = 0;
            comm_ptr->isRepresentative = 0;
        }

        if(isRep){
            pidCommarray = MPIU_Malloc(sizeof(int) * (totfgps-nonRepFGPs));
            for(i=0; i<totpids; i++){
                if(repPIDarray[i] > -1){
                    pidCommarray[numPidsInComm] = repPIDarray[i];
                    numPidsInComm++;
                }
            }
            MPIU_Assert(numPidsInComm == (totfgps-nonRepFGPs));
            MPIU_Free(repPIDarray);
        }

        FGP_DBG_PRINTF("comm_ptr->rank=%d, comm_ptr->p_rank=%d, totfgps=%d, totpids=%d, numofcoFGPs=%d, nonRepFGPs=%d, numPidsInComm=%d,  isRep=%d, comm_ptr->isRepresentative=%d\n", comm_ptr->rank, comm_ptr->p_rank, totfgps, totpids, numofcoFGPs, nonRepFGPs, numPidsInComm, isRep, comm_ptr->isRepresentative); /* _p_rank_ */

        MPI_Comm_group(comm, orig_group);
        if(isRep){
            MPI_Group_incl(*orig_group, numPidsInComm, pidCommarray, pid_group);
        }
        else{
            /* HK: In case nonRepFGPs equals zero, then *pid_group will be set to MPI_GROUP_EMPTY.
               TODO: Double-check: What effect does calling MPI_Comm_create on MPI_GROUP_EMPTY have?  */
            MPI_Group_incl(*orig_group, nonRepFGPs, nonRepPIDarray, pid_group);
        }
        /* HK: Optimizations for:
           numPidsInComm can be 1, E.g. -nfg X -n 1
           In this case there is no need to do barrier on coFGPs twice.

           nonRepFGPs can be 0, E.g. -nfg 1 -n X
           In this case *pidgroup will be set to MPI_GROUP_EMPTY

        */
        MPI_Group_incl(*orig_group, numofcoFGPs, coFGParray, coFGP_group);
        MPI_Comm_create(comm, *coFGP_group, comm_ptr->coFGP_comm);
        FGP_DBG_PRINTF("After coFGPComm comm_create\n");
        MPI_Comm_create(comm, *pid_group, comm_ptr->pid_comm);
        FGP_DBG_PRINTF("After pidComm Comm_create\n");
        MPIU_Free(coFGParray);
        if(nonRepPIDarray){
            MPIU_Free(nonRepPIDarray); 
        }
        if(isRep){
            MPIU_Free(pidCommarray);
        }
    }
    /*
      HK: coFGP_comm and pid_comm either already exist or have been created at this point.
      Doing error checks first. Calling barrier first on coFGP_comm, then on pid_comm and
      then again on coFGP_comm 
    */
    FGP_DBG_PRINTF("comm_ptr->coFGP_comm=%p, comm_ptr->pid_comm=%p\n", comm_ptr->coFGP_comm, comm_ptr->pid_comm);

    /* Convert MPI object handles to object pointers */
    MPID_Comm_get_ptr(*(comm_ptr->coFGP_comm), *coFGP_comm_dblptr);
    MPID_Comm_get_ptr(*(comm_ptr->pid_comm), *pid_comm_dblptr);
    
    /* Validate parameters and objects (post conversion) */
#   ifdef HAVE_ERROR_CHECKING
    {
        MPID_BEGIN_ERROR_CHECKS;
        {
            MPIR_ERRTEST_COMM(*(comm_ptr->coFGP_comm), mpi_errno);
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
        
        MPID_BEGIN_ERROR_CHECKS;
        {
            MPIR_ERRTEST_COMM(*(comm_ptr->pid_comm), mpi_errno);
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
        
        MPID_BEGIN_ERROR_CHECKS;
        {
            /* Validate comm_ptr */
            MPID_Comm_valid_ptr( *coFGP_comm_dblptr, mpi_errno );
            /* If *coFGP_comm_dblptr is not valid, it will be reset to null */
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
        
        MPID_BEGIN_ERROR_CHECKS;
        {                    
            MPID_Comm_valid_ptr( *pid_comm_dblptr, mpi_errno );
            /* If *pid_comm_dblptr is not valid, it will be reset to null */
            if (mpi_errno) goto fn_fail;
        }
        MPID_END_ERROR_CHECKS;
    }
#   endif /* HAVE_ERROR_CHECKING */
    
    *myrank_ptr = (( MPID_Comm *)(*coFGP_comm_dblptr))->rank;
    *comsize_ptr = (( MPID_Comm *) (*coFGP_comm_dblptr))->totprocs;
    if (comm_ptr->isRepresentative){
        MPIU_Assert(0 == *myrank_ptr); /* HK: *coFGP_comm_dblptr->fgrank=0 will be the repersentative in coFGPs */
    }
    

    /* ... end of body of routine ... */
    
 fn_exit:
    return mpi_errno;
    
 fn_fail:
    /* --BEGIN ERROR HANDLING-- */
#   ifdef HAVE_ERROR_CHECKING
    {
	mpi_errno = MPIR_Err_create_code(
                                         mpi_errno, MPIR_ERR_RECOVERABLE, FCNAME, __LINE__, MPI_ERR_OTHER, 
                                         "**mpir_create_nestedcomm", "**mpir_create_nestedcomm %C", comm);
    }
#   endif
    mpi_errno = MPIR_Err_return_comm( comm_ptr, FCNAME, mpi_errno );
    goto fn_exit;
    /* --END ERROR HANDLING-- */
    
}

/* HK: TODO IMPORTANT
   Call find_numofcoFGPs_and_pidRepresentative() function from inside MPID_Dev_comm_create_hook(comm_)
   and remove it from MPI_Barrier() above. Adding it to MPID_Dev_comm_create_hook() will allow
   information like comm_ptr->isRepresentative, comm_ptr->numofcoFGPs, comm_ptr->numPidsInComm to be
   available to other functions, apart from MPI_Barrier(), that may need it.
*/
static int find_numofcoFGPs_and_pidRepresentative( MPID_Comm *comm_ptr )
{
    MPID_Comm *world_comm_ptr = NULL;
    int nfg = -1, hwps = -1, collocated_startrank = -1;
        
    /* HK: isRepresentative, numofcoFGPs and numPidsInComm are lazily assigned */
    if( 0 == comm_ptr->numofcoFGPs ) {
        assert(0 == comm_ptr->numPidsInComm);

        /* Convert MPI object handles to object pointers */
        MPID_Comm_get_ptr( MPI_COMM_WORLD, world_comm_ptr );
        
        if (comm_ptr == world_comm_ptr) /* If the communicator is MPI_COMM_WORLD */
        { 
            MPIX_Get_collocated_size( &nfg );
            MPIX_Get_collocated_startrank( &collocated_startrank );
            
            comm_ptr->numofcoFGPs = nfg;
            comm_ptr->numPidsInComm = comm_ptr->local_size; /*HK: Total number of HWPs in MPI_COMM_WORLD */
            
            if( collocated_startrank == comm_ptr->rank ){ /*HK: If the calling FGP is the smallest rank in my pid */
                comm_ptr->isRepresentative = 1;
            }
            else{
                comm_ptr->isRepresentative = 0;
            }
            
            goto fn_exit;
        }
        
        int totfgps = comm_ptr->totprocs;
        int totpids = comm_ptr->local_size;
        int *repPIDarray = MPIU_Malloc(sizeof(int) * totpids); /* HK: pid is the index
                                                                  of this array and value
                                                                  at index is the smallest
                                                                  FGP rank in that pid */
        memset(repPIDarray, -1, sizeof(int) * totpids); /* HK: Initializing to -1 */

        int i, numofcoFGPs=0, numPidsInComm=0, nonRepFGPs=0;

        for (i=0; i<totfgps; i++){
            int pid = -1, worldrank = -1;
            RTWPmapFind(comm_ptr->co_shared_vars->rtw_map, i, &worldrank, &pid);
            if(pid == comm_ptr->p_rank){ /* HK: It is a coFGP if it has the same pid as the calling FGP. _p_rank_ */
                numofcoFGPs++;                        
            }
            FGP_DBG_PRINTF("fgrank in comm=%d, pid=%d\n", i, pid);
            /* Separating the representative and non-representative FGPs */
            if( (repPIDarray[pid] != -1) && (repPIDarray[pid] < i) ){
                nonRepFGPs++;
            }
            if( (repPIDarray[pid] == -1) || (repPIDarray[pid] > i) ){
                if(repPIDarray[pid] != -1){
                    nonRepFGPs++;
                }
                repPIDarray[pid] = i;
            }
        }

        if( repPIDarray[comm_ptr->p_rank] == comm_ptr->rank ){ /* HK: If the calling FGP is the smallest rank in my pid. _p_rank_ */
            comm_ptr->isRepresentative = 1;
        }
        else{
            comm_ptr->isRepresentative = 0;
        }
        MPIU_Free(repPIDarray);
        numPidsInComm = totfgps - nonRepFGPs;
        comm_ptr->numofcoFGPs = numofcoFGPs;
        comm_ptr->numPidsInComm = numPidsInComm;
    }

 fn_exit:
    return (MPI_SUCCESS);
}




#endif /* #ifdef FINEGRAIN_MPI */
