/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "mpidimpl.h"

#if defined(HAVE_LIMITS_H)
#include <limits.h>
#endif
#if defined(HAVE_UNISTD_H)
#include <unistd.h>
#endif

/* FIXME: This does not belong here */
#ifdef USE_MPIU_DBG_PRINT_VC
char *MPIU_DBG_parent_str = "?";
#endif

/* FIXME: the PMI init function should ONLY do the PMI operations, not the 
   process group or bc operations.  These should be in a separate routine */
#include "pmi.h"
static int InitPG( int *argc_p, char ***argv_p,
		   int *has_args, int *has_env, int *has_parent, 
		   int *pg_rank_p, MPIDI_PG_t **pg_p );
static int MPIDI_CH3I_PG_Compare_ids(void * id1, void * id2);
static int MPIDI_CH3I_PG_Destroy(MPIDI_PG_t * pg );


#if defined(FINEGRAIN_MPI)

inline void* FG_MPIU_Malloc(size_t a, const char fcname[], int lineno){
    void * retval;
    retval = malloc((size_t)(a));
    if(retval){
        return retval;
    }
    else{
        printf("%s:_Line_%d Memory Allocation Failure!\n", fcname, lineno);
        assert(0);
    }
    return NULL;
}

inline void FG_yield(const char fcname[], int lineno){
    int numfgps;
    PMI_Get_numFGPs(&numfgps);
    if( (numfgps > 1) && ((runlist_size() > 0) || (sleepq_size() > 0)) )
    { /* HK: -nfg X, where X is numfgps*/
        thread_yield();
    }
}

inline void MPIX_Yield(void){
    thread_yield();
}

void  mpix_yield_ (void){
     MPIX_Yield();
}

inline void MPIX_Usleep(unsigned long long utime)
{
    thread_usleep(utime);
}

void mpix_usleep_(unsigned long long utime)
{
     MPIX_Usleep(utime);
}

inline void FG_yield_on_event(scheduler_event yld_event, const char fcname[], int lineno){ 
    int numfgps;
    PMI_Get_numFGPs(&numfgps);
    if( (numfgps > 1) && ((runlist_size() > 0) || (sleepq_size() > 0)) )
    { /* HK: -nfg X, where X is numfgps*/
        thread_yield_on_event(yld_event);
    }
}

inline void FG_notify_on_event(int worldrank, int action, const char fcname[], int lineno){
    scheduler_event notification;
    notification.worldrank = worldrank;
    notification.action_on_event = action;
    thread_notify_on_event(notification);    
}


/* HK: Some parts of the process group, channel and net module initialization
   is to be done only once per HWP.
   Here, we record the state of initialization. The first FGP will complete
   the initialization, before other FGPs are allowed to enter InitPG() and subsequent functions. */ 
MPID_Request ** FG_recvq_posted_head = 0;
MPID_Request ** FG_recvq_posted_tail = 0;
MPID_Request ** FG_recvq_unexpected_head = 0;
MPID_Request ** FG_recvq_unexpected_tail = 0;

FGP_Init_State_t FGP_init_state = FGP_PRE_INIT;
static int All_FGP_appnum = -1; /* HK: appnum could also have been a macro like pg_id. TODO later? */
static char *All_FGP_pg_id = NULL; /* HK: #define pg_id All_FGP_pg_id */
static int All_FGP_has_parent = 0; /* HK: #define has_parent All_FGP_has_parent */
static MPIDI_PG_t *All_FGP_pg = 0; /* HK: #define pg All_FGP_pg in InitPG */
#endif
MPIDI_Process_t MPIDI_Process = { NULL }; /* HK: Once per HWP */

MPIDI_CH3U_SRBuf_element_t * MPIDI_CH3U_SRBuf_pool = NULL; /* HK: TODO global */

#undef FUNCNAME
#define FUNCNAME MPID_Init
#undef FCNAME
#define FCNAME MPIDI_QUOTE(FUNCNAME)
int MPID_Init(int *argc, char ***argv, int requested, int *provided, 
	      int *has_args, int *has_env)
{
    int mpi_errno = MPI_SUCCESS;
    
#if defined(FINEGRAIN_MPI)
#define has_parent All_FGP_has_parent
#else
    int has_parent;
#endif
    
    MPIDI_PG_t * pg=NULL;
    int pg_rank=-1;
    int pg_size;
    MPID_Comm * comm;
    int p;
   
    MPIDI_STATE_DECL(MPID_STATE_MPID_INIT);

    /* HK: Following macro expands to MPIR_FUNC_ENTER(this is for tracing and debugging)  or MPIDU_FUNC_ENTER(This accesses a global variable that keeps track of recursion level. HK: TODO fix for global variable.*/
    MPIDI_FUNC_ENTER(MPID_STATE_MPID_INIT); 

    /* FIXME: This is a good place to check for environment variables
       and command line options that may control the device */

#if 1
    /* This is a sanity check because we define a generic packet size
     */
    if (sizeof(MPIDI_CH3_PktGeneric_t) < sizeof(MPIDI_CH3_Pkt_t)) {
	fprintf( stderr, "Internal error - packet definition is too small.  Generic is %d bytes, MPIDI_CH3_Pkt_t is %d\n", sizeof(MPIDI_CH3_PktGeneric_t),
		 sizeof(MPIDI_CH3_Pkt_t) );
	exit(1);
    }
#endif

    /*
     * Set global process attributes.  These can be overridden by the channel 
     * if necessary.
     */
    MPIR_Process.attrs.tag_ub          = MPIDI_TAG_UB;

    /* If the channel requires any setup before making any other 
       channel calls (including CH3_PG_Init), the channel will define
       this routine (the dynamically loaded channel uses this) */
#ifdef HAVE_CH3_PRELOAD
#if defined(FINEGRAIN_MPI)
    /*HK: This code is not reached in my test setup
      HK: Tries to locate dll and a number of functions inside it. Has a lot of global and static vars HK: TODO later*/
    printf("``````````````````` Inside %s: ifdef HAVE_CH3_PRELOAD is true\n",__FUNCTION__);
#endif
    mpi_errno = MPIDI_CH3_PreLoad(); 
    if (mpi_errno) { MPIU_ERR_POP(mpi_errno); }
#endif
    
    /*
     * Perform channel-independent PMI initialization
     */
    /* HK: OUT parameter pg has been filled in at the return of this function in static variable All_FGP_pg */
    mpi_errno = InitPG( argc, argv, 
                        has_args, has_env, &has_parent, &pg_rank, &pg ); 
    if (mpi_errno) {
       MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, "**ch3|ch3_init");
    }
    
    
    /*
     * Let the channel perform any necessary initialization
     * The channel init should assume that PMI_Init has been called and that
     * the basic information about the job has been extracted from PMI (e.g.,
     * the size and rank of this process, and the process group id)
     */
#if defined (FINEGRAIN_MPI)
  /* HK: MPIDI_CH3_Init() is being called here */
  if(FGP_WITHIN_INIT == FGP_init_state) 
  {
#endif
    mpi_errno = MPIU_CALL(MPIDI_CH3,Init(has_parent, pg, pg_rank));
    if (mpi_errno != MPI_SUCCESS) {
	MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, "**ch3|ch3_init");
    }
#if defined (FINEGRAIN_MPI)
  }
#endif
  
    /* FIXME: Why are pg_size and pg_rank handled differently? */
    pg_size = MPIDI_PG_Get_size(pg);

#if defined (FINEGRAIN_MPI)
  if(FGP_WITHIN_INIT == FGP_init_state) 
  {
#endif    
    /* HK: MPIDI_Process is per HWP. */
    MPIDI_Process.my_pg = pg;  /* brad : this is rework for shared memories 
				* because they need this set earlier
                                * for getting the business card
                                */
    MPIDI_Process.my_pg_rank = pg_rank;

    /* FIXME: Why do we add a ref to pg here? */
    MPIDI_PG_add_ref(pg); /* HK: TODO Double-check. Apparently MPI_Finalize releases resources based on the ref_count */

#if defined (FINEGRAIN_MPI)    
    int totprocs;
    PMI_Get_totprocs(&totprocs);

    world_co_shared_vars = (Coproclet_shared_vars_t *)MPIU_Malloc(sizeof(Coproclet_shared_vars_t));
    MPIR_Comm_init_coshared_all_ref(world_co_shared_vars);

    worldcomm_rtw_map = NULL;
    worldcomm_rtw_map = (RTWmap*) RTWmapWorldCreate(totprocs); /* HK: This is an empty map */
    MPIU_Assert(worldcomm_rtw_map != NULL);
    world_co_shared_vars->rtw_map =  worldcomm_rtw_map;

    worldcomm_barrier_vars = NULL;
    worldcomm_barrier_vars = (struct coproclet_barrier_vars *)MPIU_Malloc(sizeof(struct coproclet_barrier_vars));
    MPIU_Assert(worldcomm_barrier_vars != NULL);
    worldcomm_barrier_vars->coproclet_signal = 0;
    worldcomm_barrier_vars->coproclet_counter = 0;
    worldcomm_barrier_vars->leader_signal = 0;
    world_co_shared_vars->co_barrier_vars = worldcomm_barrier_vars;
    
    contextLeader_hshtbl = NULL;
    contextLeader_hshtbl = CL_LookupHashCreate(); /* HK: This is the hash type used for context_id leader
                                                     lookup used with MPICH2's context-id generation algorithm */
    MPIU_Assert(contextLeader_hshtbl != NULL);
    cidLookup_hshtbl = NULL;
    cidLookup_hshtbl = CidLookupHashCreate();    /* HK: This is the hash type used for context_id lookup used
                                                    with CID = <LID,LBI> context-id generation algorithm */
    MPIU_Assert(cidLookup_hshtbl != NULL);
  }
#endif

    /*
     * Initialize the MPI_COMM_WORLD object
     */
    comm = MPIR_Process.comm_world;

#if defined(FINEGRAIN_MPI)
    comm->p_rank      = pg_rank; /* HK: _p_rank_ */
#else
    comm->rank        = pg_rank; 
#endif
    comm->remote_size = pg_size;
    comm->local_size  = pg_size;
#if defined(FINEGRAIN_MPI)
    comm->rank = my_fgrank;
    PMI_Get_totprocs(&(comm->totprocs));
    comm->co_shared_vars = world_co_shared_vars;
    comm->coFGP_comm = NULL;
    comm->pid_comm = NULL;
    comm->isRepresentative = 0;
    comm->numofcoFGPs = 0;
    comm->numPidsInComm = 0;        
    comm->barrier_vars = NULL;
#endif

#if defined(FINEGRAIN_MPI)
    if(FGP_WITHIN_INIT == FGP_init_state) 
    {
        mpi_errno = MPID_VCRT_Create(comm->remote_size, &vcrt_world); /* HK: Sets ref_count to 1 TODO: Double-check */
        if (mpi_errno != MPI_SUCCESS)
        {
            MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,"**dev|vcrt_create", 
                                 "**dev|vcrt_create %s", "MPI_COMM_WORLD");
        }
        
        mpi_errno = MPID_VCRT_Get_ptr(vcrt_world, &vcr_world); /* HK: sets comm->vcr = (comm->vcrt)->vcr_table; */
        if (mpi_errno != MPI_SUCCESS)
        {
                MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,"**dev|vcrt_get_ptr", 
                                     "dev|vcrt_get_ptr %s", "MPI_COMM_WORLD");
        }
        
        /* Initialize the connection table on COMM_WORLD from the process group's
           connection table */
        for (p = 0; p < pg_size; p++)
        {
            MPID_VCR_Dup(&pg->vct[p], &vcr_world[p]);
        }
    }
    MPIU_Assert( (vcrt_world != NULL) && (vcr_world != NULL) );
    comm->vcrt = vcrt_world;
    comm->vcr  = vcr_world;
    if(FGP_WITHIN_INIT == FGP_init_state) {
        MPID_Dev_comm_create_hook (comm); /* HK: This resolves to MPIDI_CH3I_comm_create() */
        world_ch3i_ptr = &(comm->ch);        
    }
    else{
        MPIR_Copy_ch3i_comm_ch(comm, world_ch3i_ptr);
    }
    world_co_shared_vars->ch3i_ptr = world_ch3i_ptr;

#else
    mpi_errno = MPID_VCRT_Create(comm->remote_size, &comm->vcrt); /* HK: Sets ref_count to 1 TODO: Double-check */
    if (mpi_errno != MPI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,"**dev|vcrt_create", 
			     "**dev|vcrt_create %s", "MPI_COMM_WORLD");
    }
    
    mpi_errno = MPID_VCRT_Get_ptr(comm->vcrt, &comm->vcr);
    if (mpi_errno != MPI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,"**dev|vcrt_get_ptr", 
			     "dev|vcrt_get_ptr %s", "MPI_COMM_WORLD");
    }
    
    /* Initialize the connection table on COMM_WORLD from the process group's
       connection table */
    for (p = 0; p < pg_size; p++)
    {
	MPID_VCR_Dup(&pg->vct[p], &comm->vcr[p]);
    }

    MPID_Dev_comm_create_hook (comm); /* HK: This resolves to MPIDI_CH3I_comm_create() */
#endif
    
    /*
     * Initialize the MPI_COMM_SELF object
     */
    /* HK: Adding comm->fgrank and comm->totprocs here TODO: Double-check
       The reason is that MPI_Isend gets called along the MPI_Finalize path and
       in MPID_Send with destination rank=0 and source rank=1 was resulting in send_to_self.
       This was because a communicator other than comm_world is being used in MPI_Finalize and
       in COMPARE_RANKS, which compares destination rank=0 with comm->fgrank, a match was
       erroneously occuring because comm_fgrank was not set properly and originally contained zero.
       Setting comm->fgrank here to fix this problem. MPI_Finalize is
       no longer looping endlessly in the progress engine for sndrecv.c
    */
    comm = MPIR_Process.comm_self;
#if defined(FINEGRAIN_MPI)
    comm->p_rank      = 0;
#else
    comm->rank        = 0;
#endif
    comm->remote_size = 1;
    comm->local_size  = 1;
#if defined(FINEGRAIN_MPI)
    comm->rank = 0; /* HK: TODO Double-check */
    comm->totprocs = 1; /* HK: TODO Double-check */
    comm->co_shared_vars = world_co_shared_vars; /* HK: TODO Double-check */
    comm->coFGP_comm = NULL;
    comm->pid_comm = NULL;
    comm->isRepresentative = 0;
    comm->numofcoFGPs = 0;
    comm->numPidsInComm = 0;        
    comm->barrier_vars = NULL;
#endif

    
    mpi_errno = MPID_VCRT_Create(comm->remote_size, &comm->vcrt);
    if (mpi_errno != MPI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, "**dev|vcrt_create", 
			     "**dev|vcrt_create %s", "MPI_COMM_SELF");
    }
    
    mpi_errno = MPID_VCRT_Get_ptr(comm->vcrt, &comm->vcr);
    if (mpi_errno != MPI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, "**dev|vcrt_get_ptr", 
			     "dev|vcrt_get_ptr %s", "MPI_COMM_WORLD");
    }
    
    MPID_VCR_Dup(&pg->vct[pg_rank], &comm->vcr[0]);

    /* Currently, mpidpre.h always defines MPID_NEEDS_ICOMM_WORLD. */
#ifdef MPID_NEEDS_ICOMM_WORLD
    /*
     * Initialize the MPIR_ICOMM_WORLD object (an internal, private version
     * of MPI_COMM_WORLD) 
     */
    comm = MPIR_Process.icomm_world;
    
    /* HK: Adding comm->fgrank and comm->totprocs here. TODO: Double-check
       See comment above.
    */
#if defined(FINEGRAIN_MPI)
    comm->p_rank      = pg_rank; /* HK: _p_rank_ */
#else
    comm->rank        = pg_rank;
#endif
    comm->remote_size = pg_size;
    comm->local_size  = pg_size;
#if defined(FINEGRAIN_MPI)
    comm->rank = my_fgrank;
    PMI_Get_totprocs(&(comm->totprocs));
    comm->co_shared_vars = world_co_shared_vars; 
    comm->coFGP_comm = NULL;
    comm->pid_comm = NULL;
    comm->isRepresentative = 0;
    comm->numofcoFGPs = 0;
    comm->numPidsInComm = 0;        
    comm->barrier_vars = NULL;
#endif

#if 0    
    mpi_errno = MPID_VCRT_Create(comm->remote_size, &comm->vcrt);
    if (mpi_errno != MPI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,"**dev|vcrt_create", 
			     "**dev|vcrt_create %s", "MPI_COMM_WORLD");
    }
    
    mpi_errno = MPID_VCRT_Get_ptr(comm->vcrt, &comm->vcr);
    if (mpi_errno != MPI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,"**dev|vcrt_get_ptr", 
			     "dev|vcrt_get_ptr %s", "MPI_COMM_WORLD");
    }
    
    /* Initialize the connection table on COMM_WORLD from the process group's
       connection table */
    for (p = 0; p < pg_size; p++)
    {
	MPID_VCR_Dup(&pg->vct[p], &comm->vcr[p]);
    }
#endif
    MPID_VCRT_Add_ref( MPIR_Process.comm_world->vcrt ); /* HK: TODO Double-check. This should be fine as MPIR_Process.comm_world is per FGP */
    comm->vcrt = MPIR_Process.comm_world->vcrt;
    comm->vcr  = MPIR_Process.comm_world->vcr;
#if defined(FINEGRAIN_MPI)
    MPIR_Copy_ch3i_comm_ch(comm, world_ch3i_ptr);
#else
    MPID_Dev_comm_create_hook (comm);
#endif
#endif


#if defined(FINEGRAIN_MPI) /* HK: Adding the predefined context-ids of
                              communicators such as MPI_COMM_WORLD (context-id=0),
                              MPI_COMM_SELF (context-id=4), MPIR_ICOMM_WORLD (context-id=8)
                              to cidLookup_hshtbl
                           */
    if(FGP_WITHIN_INIT == FGP_init_state) 
    {
        cidLookupHashItemptr stored = NULL;
        comm = MPIR_Process.comm_world;
        CidLookupHashInsert(cidLookup_hshtbl, comm->context_id, world_co_shared_vars, &stored);
        assert(stored);
        stored = NULL;
        comm = MPIR_Process.comm_self;
        CidLookupHashInsert(cidLookup_hshtbl, comm->context_id, world_co_shared_vars, &stored);
        assert(stored);
        stored = NULL;
        comm = MPIR_Process.icomm_world;
        CidLookupHashInsert(cidLookup_hshtbl, comm->context_id, world_co_shared_vars, &stored);
        assert(stored);

        /* HK: Creating an array of pre-posted receive requests */
        int numfgps;
        PMI_Get_numFGPs(&numfgps); 
        FG_recvq_posted_head = (MPID_Request **)MPIU_Calloc(numfgps, sizeof(MPID_Request *));
        FG_recvq_posted_tail = (MPID_Request **)MPIU_Calloc(numfgps, sizeof(MPID_Request *));
        FG_recvq_unexpected_head = (MPID_Request **)MPIU_Calloc(numfgps, sizeof(MPID_Request *));
        FG_recvq_unexpected_tail = (MPID_Request **)MPIU_Calloc(numfgps, sizeof(MPID_Request *));

    }
#endif

    
    /*
     * If this process group was spawned by a MPI application, then
     * form the MPI_COMM_PARENT inter-communicator.
     */

    /*
     * FIXME: The code to handle the parent case should be in a separate 
     * routine and should not rely on #ifdefs
     */
#ifndef MPIDI_CH3_HAS_NO_DYNAMIC_PROCESS   
    if (has_parent) {
	char * parent_port;

	/* FIXME: To allow just the "root" process to 
	   request the port and then use MPIR_Bcast to 
	   distribute it to the rest of the processes,
	   we need to perform the Bcast after MPI is
	   otherwise initialized.  We could do this
	   by adding another MPID call that the MPI_Init(_thread)
	   routine would make after the rest of MPI is 
	   initialized, but before MPI_Init returns.
	   In fact, such a routine could be used to 
	   perform various checks, including parameter
	   consistency value (e.g., all processes have the
	   same environment variable values). Alternately,
	   we could allow a few routines to operate with 
	   predefined parameter choices (e.g., bcast, allreduce)
	   for the purposes of initialization. */
	mpi_errno = MPIDI_CH3_GetParentPort(&parent_port);
	if (mpi_errno != MPI_SUCCESS) {
	    MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, 
				"**ch3|get_parent_port");
	}
	MPIU_DBG_MSG_S(CH3_CONNECT,VERBOSE,"Parent port is %s\n", parent_port);
	    
	mpi_errno = MPID_Comm_connect(parent_port, NULL, 0, 
				      MPIR_Process.comm_world, &comm);
	if (mpi_errno != MPI_SUCCESS) {
	    MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,
				 "**ch3|conn_parent", 
				 "**ch3|conn_parent %s", parent_port);
	}

	MPIR_Process.comm_parent = comm;
	MPIU_Assert(MPIR_Process.comm_parent != NULL);
	MPIU_Strncpy(comm->name, "MPI_COMM_PARENT", MPI_MAX_OBJECT_NAME);
	
	/* FIXME: Check that this intercommunicator gets freed in MPI_Finalize
	   if not already freed.  */
        
#if defined(FINEGRAIN_MPI)
        printf("``````````````````` In %s: pg_rank is %d and inside #ifndef MPIDI_CH3_HAS_NO_DYNAMIC_PROCESS and if(has_parent)\n", __FUNCTION__, pg_rank);
#endif
    }
#endif	
    
    /*
     * Set provided thread level
     */
    if (provided != NULL)
    {
	/* This must be min(requested,MPICH_THREAD_LEVEL) if runtime
	   control of thread level is available */
	*provided = (MPICH_THREAD_LEVEL < requested) ? 
	    MPICH_THREAD_LEVEL : requested;
    }

  fn_exit:
    MPIDI_FUNC_EXIT(MPID_STATE_MPID_INIT);
    return mpi_errno;

    /* --BEGIN ERROR HANDLING-- */
  fn_fail:
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}

/* This allows each channel to perform final initialization after the
 rest of MPI_Init completes.  */
int MPID_InitCompleted( void )
{
    int mpi_errno;
    mpi_errno = MPIU_CALL(MPIDI_CH3,InitCompleted());
    return mpi_errno;
}

/*
 * Initialize the process group structure by using PMI calls.
 * This routine initializes PMI and uses PMI calls to setup the 
 * process group structures.
 * 
 */
static int InitPG( int *argc, char ***argv, 
		   int *has_args, int *has_env, int *has_parent, 
		   int *pg_rank_p, MPIDI_PG_t **pg_p )
{
    int pmi_errno;
    int mpi_errno = MPI_SUCCESS;
    int pg_rank, pg_size, appnum, pg_id_sz;
    int usePMI=1;
#if defined (FINEGRAIN_MPI)
#define pg_id  All_FGP_pg_id
#define pg All_FGP_pg
#else
    char *pg_id;
    MPIDI_PG_t *pg = 0;
#endif


    /* See if the channel will provide the PMI values.  The channel
     is responsible for defining HAVE_CH3_PRE_INIT and providing 
    the MPIDI_CH3_Pre_init function.  */
    /* FIXME: Document this */
#ifdef HAVE_CH3_PRE_INIT
    {
#if defined (FINEGRAIN_MPI)
        /* HK: This code is not reached in my test setup. MPIDI_CH3_Pre_init does gasnet initialization. HK: TODO later */
        printf("``````````````````` Inside %s: #ifdef HAVE_CH3_PRE_INIT is true\n",__FUNCTION__);
#endif
	int setvals;
	mpi_errno = MPIDI_CH3_Pre_init( &setvals, has_parent, &pg_rank, 
					&pg_size );
	if (mpi_errno) {
	    goto fn_fail;
	}
	if (setvals) usePMI = 0;
    }
#endif 

    /* If we use PMI here, make the PMI calls to get the
       basic values.  Note that systems that return setvals == true
       do not make use of PMI for the KVS routines either (it is
       assumed that the discover connection information through some
       other mechanism */
    /* FIXME: We may want to allow the channel to ifdef out the use
       of PMI calls, or ask the channel to provide stubs that 
       return errors if the routines are in fact used */
    if (usePMI) {
	/*
	 * Initialize the process manangement interface (PMI), 
	 * and get rank and size information about our process group
	 */
#if defined (FINEGRAIN_MPI)
     if(FGP_WITHIN_INIT == FGP_init_state) /* HK: OUT parameter has_parent is defined as All_FGP_has_parent */
     {
#endif
	pmi_errno = PMI_Init(has_parent);

	if (pmi_errno != PMI_SUCCESS) {
	    MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, "**pmi_init",
			     "**pmi_init %d", pmi_errno);
	}
#if defined (FINEGRAIN_MPI)
     }

     MPIU_Assert(curr_fgrank >= 0); 
     my_fgrank = curr_fgrank; 
     curr_fgrank++;
#endif

	pmi_errno = PMI_Get_rank(&pg_rank); /* HK:  sets pg_rank = PMI_rank; */
	if (pmi_errno != PMI_SUCCESS) {
	    MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, "**pmi_get_rank",
			     "**pmi_get_rank %d", pmi_errno);
	}

	pmi_errno = PMI_Get_size(&pg_size); /* HK:  sets pg_size = PMI_size; */
	if (pmi_errno != 0) {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, "**pmi_get_size",
			     "**pmi_get_size %d", pmi_errno);
	}
	/* HK: 
           PMI_Get_appnum() talks to pm server through MPI_fd. It sends a command
           "cmd = get_appnum". In the python script mpdman.py, handle_pmi_input() returns
           self.appnum. appnum is global and is being appnum+=1 in handli_local_argset() in mpiexec.py.
           Below, MPIR_Process.attrs.appnum = appnum; so all FGPs will have same appnum. HK: TODO implications?
         */
#if defined (FINEGRAIN_MPI)
     if(FGP_WITHIN_INIT == FGP_init_state) /* HK: OUT parameter appnum is being stored in All_FGP_appnum */
     {
#endif
	pmi_errno = PMI_Get_appnum(&appnum);
	if (pmi_errno != PMI_SUCCESS) {
	    MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, "**pmi_get_appnum",
				 "**pmi_get_appnum %d", pmi_errno);
	}
#if defined (FINEGRAIN_MPI)
        All_FGP_appnum = appnum;   
     }
#endif     


#if defined (FINEGRAIN_MPI)
     if (All_FGP_appnum != -1) {
         appnum = All_FGP_appnum; /* HK: Have to do this assignment for FGPs coming after the first one. */
#else
         /* Note that if pmi is not availble, the value of MPI_APPNUM is 
            not set */     
         if (appnum != -1) {
#endif
          
              MPIR_Process.attrs.appnum = appnum;
          }
	
	/* Now, initialize the process group information with PMI calls */
	/*
	 * Get the process group id
	 */
	pmi_errno = PMI_Get_id_length_max(&pg_id_sz); /* HK: sets pg_id_sz = PMI_kvsname_max; */
	if (pmi_errno != PMI_SUCCESS) {
	    MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER,
				 "**pmi_get_id_length_max", 
				 "**pmi_get_id_length_max %d", pmi_errno);
	}

#if defined (FINEGRAIN_MPI)
      if(FGP_WITHIN_INIT == FGP_init_state) 
      {
#endif        
	/* This memory will be freed by the PG_Destroy if there is an error */
	pg_id = MPIU_Malloc(pg_id_sz + 1);
	if (pg_id == NULL) {
	    MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, "**nomem");
	}

	/* Note in the singleton init case, the pg_id is a dummy.
	   We'll want to replace this value if we join an 
	   Process manager */
	pmi_errno = PMI_Get_id(pg_id, pg_id_sz);
	if (pmi_errno != PMI_SUCCESS) {
	    MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, "**pmi_get_id",
				 "**pmi_get_id %d", pmi_errno);
	}
#if defined (FINEGRAIN_MPI)
      }
#endif
     }
    else {
	/* Create a default pg id */
	pg_id = MPIU_Malloc(2);
	if (pg_id == NULL) {
	    MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, "**nomem");
	}
	MPIU_Strncpy( pg_id, "0", 2 );
    }

     /* HK: No need to call MPIDI_PG_Init() per FGP. It is only setting the function pointers in mpidi_pg.c
        static MPIDI_PG_Compare_ids_fn_t MPIDI_PG_Compare_ids_fn;
        static MPIDI_PG_Destroy_fn_t MPIDI_PG_Destroy_fn;
        to MPIDI_CH3I_PG_Compare_ids and MPIDI_CH3I_PG_Destroy respectively.
        It also does getenv(MPICHD_DBG_PG) HK: TODO later?

        Calling MPIDI_PG_Create(pg_size, pg_id, &pg); and MPIDI_PG_InitConnKVS( pg );  once per HWP.
        MPIDI_PG_Create() will also initialize the VC Table.
        
     */
#if defined (FINEGRAIN_MPI)
   if(FGP_WITHIN_INIT == FGP_init_state)
   {
#endif        

    /*
     * Initialize the process group tracking subsystem
     */
    mpi_errno = MPIDI_PG_Init(argc, argv, 
                              MPIDI_CH3I_PG_Compare_ids, MPIDI_CH3I_PG_Destroy);
    if (mpi_errno != MPI_SUCCESS) {
	MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER,"**dev|pg_init");
    }
    

    /*
     * Create a new structure to track the process group for our MPI_COMM_WORLD
     */
   mpi_errno = MPIDI_PG_Create(pg_size, pg_id, &pg); /* HK: Will be called by first FGP only. */
    if (mpi_errno != MPI_SUCCESS) {
	MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, "**dev|pg_create");
    }

    /* FIXME: We can allow the channels to tell the PG how to get
       connection information by passing the pg to the channel init routine */
    if (usePMI) {
	/* Tell the process group how to get connection information */
	MPIDI_PG_InitConnKVS( pg );

#if 0
	/* If we're supporting the debugger, we can save our host and 
	   pid here.  This publishes the data in the kvs space. 
	   This allows the MPI processes to access the information 
	   about the other processes without any PMI changes. */
#ifdef HAVE_DEBUGGER_SUPPORT
	{
	    char key[64];
	    char myinfo[512];
	    MPIU_Snprintf( key, sizeof(key), "hpid-%d", pg_rank );
	    MPIU_Snpritnf( myinfo, sizeof(key), "%s:%d", hostname, getpid() );
	    PMI_KVS_Put( pg->world->connData, key, myinfo );
	}
#endif
#endif
    }
#if defined (FINEGRAIN_MPI)
   }
#endif

   
    /* FIXME: Who is this for and where does it belong? */
#ifdef USE_MPIU_DBG_PRINT_VC
    MPIU_DBG_parent_str = (*has_parent) ? "+" : "";
#endif

    /* FIXME: has_args and has_env need to come from PMI eventually... */
    *has_args = TRUE;
    *has_env  = TRUE;

    *pg_p      = pg;
    *pg_rank_p = pg_rank;
    
 fn_exit:
    return mpi_errno;
 fn_fail:
    /* --BEGIN ERROR HANDLING-- */
    if (pg) {
	MPIDI_PG_Destroy( pg );
    }
    goto fn_exit;
    /* --END ERROR HANDLING-- */
}

/*
 * Create the storage for the business card. 
 *
 * The routine MPIDI_CH3I_BCFree should be called with the original 
 * value *bc_val_p .  Note that the routines that set the value 
 * of the businesscard return a pointer to the first free location,
 * so you need to remember the original location in order to free 
 * it later.
 *
 */
int MPIDI_CH3I_BCInit( char **bc_val_p, int *val_max_sz_p )
{
    int pmi_errno;
    int mpi_errno = MPI_SUCCESS;

    pmi_errno = PMI_KVS_Get_value_length_max(val_max_sz_p);
    if (pmi_errno != PMI_SUCCESS)
    {
	MPIU_ERR_SETANDJUMP1(mpi_errno,MPI_ERR_OTHER, 
			     "**pmi_kvs_get_value_length_max",
			     "**pmi_kvs_get_value_length_max %d", pmi_errno);
    }

    /* This memroy is returned by this routine */
    *bc_val_p = MPIU_Malloc(*val_max_sz_p);
    if (*bc_val_p == NULL) {
	MPIU_ERR_SETANDJUMP(mpi_errno,MPI_ERR_OTHER, "**nomem");
    }
    
    /* Add a null to simplify looking at the bc */
    **bc_val_p = 0;

  fn_exit:
    return mpi_errno;

  fn_fail:
    goto fn_exit;
}

/* Free the business card.  This routine should be called once the business
   card is published. */
int MPIDI_CH3I_BCFree( char *bc_val )
{
    /* */
    if (bc_val) {
	MPIU_Free( bc_val );
    }
    
    return 0;
}

/* FIXME: The PG code should supply these, since it knows how the 
   pg_ids and other data are represented */
static int MPIDI_CH3I_PG_Compare_ids(void * id1, void * id2)
{
    return (strcmp((char *) id1, (char *) id2) == 0) ? TRUE : FALSE;
}


static int MPIDI_CH3I_PG_Destroy(MPIDI_PG_t * pg)
{
    if (pg->id != NULL)
    { 
	MPIU_Free(pg->id);
    }
    
    return MPI_SUCCESS;
}
